#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://trac.edgewall.com/license.html.
#

from trac.core import implements, Component
from trac.wiki import IWikiSyntaxProvider, IWikiMacroProvider

import os
import sha
import re

class Formula(Component):

    implements(IWikiSyntaxProvider, IWikiMacroProvider)
    
    outputVersion = "0.1"

    def __init__(self):
	env = self.env
        # gets paths from configuration
	self.tmpdir = env.config.get('latex', 'temp_dir')

        self.fleqnIndent = env.config.get('latex', 'fleqn_indent')
	self.latexPath = env.config.get('latex', 'latex_path')
        self.dvipsPath = env.config.get('latex', 'dvips_path')
        self.convertPath = env.config.get('latex', 'convert_path')
        self.texMag = env.config.get('latex', 'text_mag')
        self.imageFormat = env.config.get('latex', 'image_format')
    
        self.imagePath = os.path.normpath(os.path.join(env.get_htdocs_dir(), "formulas"))
        if not os.path.exists(self.imagePath):
            try:
                os.mkdir(self.imagePath)
            except:
                return "<b>Error: unable to create image directory</b><br>"

        if not self.tmpdir:
            self.tmpdir='/tmp'
        # set defaults
        if not self.fleqnIndent:
            self.fleqnIndent = '5%'
        if not self.latexPath:
            self.latexPath = 'latex'
        if not self.dvipsPath:
            self.dvipsPath = 'dvips'
        if not self.convertPath:
            self.convertPath = 'convert'
        if not self.texMag:
            self.texMag = 1000 # I'm told this is latex's default value
        if not self.imageFormat:
            self.imageFormat = 'png'

        path = self.tmpdir
        # create temporary directory if necessary
        def mkd(path):
    	    if not os.path.exists(path):
	        d, t, = os.path.split(path)
    	    if not os.path.exists(d):
    	        mkd(d)            
            os.mkdir(path)
                 
        try:
	    if not os.path.exists(path):
    	        mkd(path)
        except:
	    return "Unable to create temporary directory " + path

    def render(self, texData, density, fleqnMode, mathMode):    
        # generate final image name.  Use a hash of the parameters which affect
	# the image, so we don't have to recreate it unless they change.
        hash = sha.new(texData)
        # include some options in the hash, as they affect the output image
	hash.update( "%d %d" % (density, int(self.texMag)) ) 
        hash.update( self.outputVersion )
	name = hash.hexdigest()
        imageFile = "%s/%s.%s" % (self.imagePath, name, self.imageFormat)

        log = "<br/>"
	if not os.path.exists(imageFile):
    	    # latex writes out lots of stuff to the current directory, so we have
	    # to run it from there.
	    cwd = os.getcwd()
	    os.chdir(self.tmpdir)

            texFile = "%s/%s.tex" % (self.tmpdir, name)
	    self.makeTexFile(texFile, texData, mathMode, self.texMag)

	    # the output from latex on stdout seems to cause problems, so sent it
	    # to /dev/null
	    cmd = "%s %s > /dev/null" % (self.latexPath, texFile)
	    os.system(cmd)
	    log += cmd + "<br/>"
	    os.chdir(cwd)

            # use dvips to convert to eps
	    dviFile = "%s/%s.dvi" % (self.tmpdir, name)
    	    epsFile = "%s/%s.eps" % (self.tmpdir, name)
    	    cmd = "%s -q -D 600 -E -n 1 -p 1 -o %s %s" % (self.dvipsPath, epsFile, dviFile)
	    os.system( cmd )
    	    log += cmd + "<br/>"

            # and finally, ImageMagick to convert from eps to [imageFormat] type
	    cmd = "%s -antialias -density %ix%i %s %s" % (self.convertPath, density, density, epsFile, imageFile)
	    os.system( cmd )
    	    log += cmd + "<br/>"

	    # Remove working files
	    auxFile = "%s/%s.aux" % (self.tmpdir, name)
    	    logFile = "%s/%s.log" % (self.tmpdir, name)
	    cmd = "rm '%s' '%s' '%s' '%s' '%s'" % (texFile, dviFile, epsFile, auxFile, logFile)
	    os.system(cmd)
	    log += cmd + "<br/>"
	    
	    os.chdir(cwd)

        if fleqnMode:
	    margin = " margin-left: %s" % self.fleqnIndent
        else:
	    margin = ""
        
        html = "<img src='%s' border='0' style='vertical-align: middle;%s' alt='%s'/>" % (self.env.href.chrome('site','formulas/%s.%s'%(name, self.imageFormat)), margin, texData)
	return html

    def makeTexFile(self, texFile, texData, mathMode, texMag):
        tex = "\\batchmode\n"
	tex += "\\documentclass{article}\n"
        tex += "\\usepackage{amsmath}\n"
        tex += "\\usepackage{amssymb}\n"
	tex += "\\usepackage{epsfig}\n"
	tex += "\\usepackage[utf8]{inputenc}\n"
        tex += "\\pagestyle{empty}\n"
        tex += "\\mag=%s\n" % texMag
        # start the document
        tex += "\\begin{document}\n"
        if mathMode:
            tex += "$$\n"
        tex += "%s\n" % texData
        if mathMode:
                tex += "$$\n"
        tex += "\\pagebreak\n"
        tex += "\\end{document}\n"
            
        FILE = open(texFile, "w")
        FILE.write( tex )
        FILE.close()
    
    # IWikiSyntaxProvider methods

    def get_wiki_syntax(self):
        yield (r"\$(?P<formula>.*)\$", lambda x, y, z: self.render_syntax(y))

    def get_link_resolvers(self):
        return []

    def render_syntax(self, y):
        return self.render(y, 100, 0, 0) 

    # IWikiMacroProvider methods

    def get_macros(self):
        yield 'formula'

    def get_macro_description(self, name):
	return ("Formula macro that allows you to include latex rendered"
		"in your page.")
	
    def render_macro(self, req, name, content):
        # defaults
        density = 100
        mathMode = 1    # default to using display-math mode for LaTeX processing
        displayMode = 0 # default to generating inline formula
        fleqnMode   = self.env.config.get('latex', 'fleqn')
        centerImage = 0
        indentImage = 0
        indentClass = ""
    
        # find some number of arguments, followed by the formula
        command = re.compile('^\s*#([^=]+)=?(.*)')
        formula = ""
        errors = ""
        for line in content.split("\n"):
            m = command.match(line)
            if m:
                if m.group(1) == "density":
                    density = int(m.group(2))
                elif m.group(1) == "nomode":
                    mathMode = 0
                elif m.group(1) == "center":
                    centerImage = 1
                    fleqnMode   = 0
                elif m.group(1) == "indent":
                    indentImage = 1
                    indentClass = m.group(2)
                elif m.group(1) == "display":
                    displayMode = 1
                elif m.group(1) == "fleqn":
                    displayMode = 1
                    fleqnMode   = 1
                else:
                    errors = '<br/>Unknown <i>formula</i> command "%s"<br/>' % m.group(1)
            else:
                formula += line + "\n"
    
        # Set display and fleqn defaults
        if displayMode:
            if fleqnMode:
                centerImage = 0
            else:
                centerImage = 1
    
        # Render formula
        format = '%s'
        if centerImage:
            format = '<center>%s</center>' % format
        if indentImage:
            if indentClass:
                format = '<p class="%s">%s</p>' % (indentClass, format)
            else:
                format = '<p>%s</p>' % format
        
        result = errors + self.render(formula, density, fleqnMode, mathMode) 
        return format % result
